function varargout = RTR_SR1(fns,params)
% Solver of Riemannian trust region with symmetric rank-1 update method
%
% INPUT:
% fns : a struct that contains required function handles
% required:
%     fns.f(x) : return objective function value at x.
%     fns.gf(x) : return the gradient of objection function at x.
% not required if params.manifold, params.retraction and params.vector_transport is provided.
%     fns.inpro(x, v1, v2) : return the inner product g_x(v1, v2) of two tangent vectors v1 and v2 on T_x M.
%     fns.proj(x, eta) : return P_x(eta) by projecting v to tangent space of x.
%     fns.Hv(x, H, v) : return tangent vector which is given by that operator H affect on tangent vector v on T_x M.
%     fns.STV(a, v) : return a tangent vector which is a times v.
%     fns.TVaddTV(v1, v2) : return a tangent vector which is equal to v1 adds v2.
%     fns.R(x, eta) : return R_x(eta), where R is a retraction, x is an element on the manifold and eta is a tangent vector of T_x M.
%     fns.invTranv(x1, d, x2, v) : return a tangent vector on T_x1 M, which is given by inverse vector transport that transport v \in T_x2 M to T_x1 M. Here x2 = R_x1(d).
%     fns.TranH(x1, d, x2, H1) : return a operator, H2, that affects on T_x2 M. H2 satisfies that for any v \in T_x2(M), H2(v) = Tran (H1 (Tran^{-1}(v)))
%                             where Tran is the vector transport fns.Tranv(x1, d, x2, v).
%     fns.rank1operator(x, v1, v2) : return a operator, H, on T_x M that satisfies that for any v \in T_x M, H(v) = g_x(v2, v) * v1.
%     fns.operadd(x, H1, H2) : return a operator, H, on T_x M that satisfies that for any v \in T_x M, H(v) = H1(v) + H2(v).
% for partly smooth function
%     fns.Tranv(x1, d, x2, v) : return a tangent vector on T_x2 M, which is given by vector transport that transport v \in T_x M to T_{R_x(d)} M. Here x2 = R_x(d).
% 
% params : a struct that contains parameters that are used in the solver.
% required:
%     params.x0 : initial approximation of minimizer.
%     params.B0 : initial approximation of Hessian.
% not required:
%     params.error [1e-5]       : tolerance of stopping criterion.
%     params.StopCriterion [3]  : stopping criterion, 0 means call fns.IsStopped to check
%                                 1 means stop when relative error of objective function is less than tolerance,
%                                 2 means stop when norm of gradient is less than tolerance,
%                                 3 means stop when norm of gradient over initial norm of gradient is less than tolerance,
%                                 4 means stopping criterion for partly smooth function.
%     params.max_t [300]        : the maximum number of iterations.
%     params.debug [1]          : '0' means almost silence, '1' means some output. '2' means a lot of output, '3' means more than need to know.
%     params.c [0.1]            : acceptance or rejection constant
%     params.tau1 [0.25]        : shrink radius constant
%     params.tau2 [2]           : expansion radius constant
%     params.max_Delta [200]    : max of Delta
%     params.min_Delta [1e-10]   : min of Delta
%     params.err_x [1e-5]       : tolerance for checking whether iterates are close to convergence. It is between [0, 1].
%     params.num_Grads [10]     : number of gradients for checking whether iterates are close to Clark stationary point. It is between [0, inf].
%     params.theta[0.1]         : A stopping criterion parameter for solving local model
%     params.kappa[0.9]         : A stopping criterion parameter for solving local model
%     params.min_innit[0]       : the minimum number of iterations for solving local model
%     params.max_innit[200]     : the maxmum number of iterations for solving local model
% The following give some examples of manifold. For each manifold, some retractions and vector transports are provided.
%     params.manifold : the manifold which objective function is on.
%         '1', sphere : S^{n - 1}
%         '2', Stiefel manifold : St(p, n)
%         '3', orthogonal group : O(n)
%         '4', grassmann manifold : Gr(p, n)
%
% OUTPUT:
% varargout : contains output parameters
%     varargout{1} : all iterates
%     varargout{2} : function values of all iterates
%     varargout{3} : norm of gradient of all iterates
%     varargout{4} : total computational time
%     varargout{5} : number of function evaluations
%     varargout{6} : number of gradient evaluations
%     varargout{7} : number of Hessian actions
%     varargout{8} : accumulated computational time for iterations
%     varargout{9} : number of vector transport actions
%     varargout{10} : number of retraction actions
% 
% By Wen Huang, the solver of local model is a modified version of that in http://www.math.fsu.edu/~cbaker/GenRTR/

    fprintf('RTR_SR1\n')
    if nargin < 2,
       error('Invalid arguments: the number of arguments is invalid.');
    end
    % set default parameters and check the legal range of parameters
    params = set_default_para(params, 'StopCriterion', 3, 'int', 0, 4);
    params = set_default_para(params, 'error', 1e-6, 'float', 0, 1);
    params = set_default_para(params, 'err_x', 1e-5, 'float', 0, 1);
    params = set_default_para(params, 'num_Grads', 10, 'int', 0, inf);
    params = set_default_para(params, 'max_t', 300, 'int', 1, inf);
    params = set_default_para(params, 'debug', 1, 'int', 0, 3);
    params = set_default_para(params, 'c', 0.1, 'float', 0, 0.25);
    params = set_default_para(params, 'tau1', 0.25, 'float', 0.1, 0.9);
    params = set_default_para(params, 'tau2', 2, 'float', 1.5, 5);
    params = set_default_para(params, 'max_Delta', 200, 'float', 1, inf);
    params = set_default_para(params, 'min_Delta', 1e-10, 'float', 0, 1);
    params = set_default_para(params, 'theta', 0.1, 'float', 0, 1);
    params = set_default_para(params, 'kappa', 0.9, 'float', 0, 1);
    params = set_default_para(params, 'min_innit', 0, 'int', 0, inf);
    params = set_default_para(params, 'max_innit', 200, 'int', 0, inf);
    params = set_default_para(params, 'isconvex', 0, 'int', 0, 1);
    params = set_default_para(params, 'keepnumX', 2100000000, 'int', 1, inf);
    params = set_default_para(params, 'useRand', 0, 'int', 0, 1);
    params = set_default_para(params, 'isplot', 0, 'int', 0, 1);
    if(isfield(params, 'manifold') && isfield(params, 'retraction') && isfield(params, 'vector_transport'))
        fs = choose_manifold(params.manifold, params.retraction, params.vector_transport);
        fns.Hv = fs.Hv;
        fns.STV = fs.STV;
        fns.TVaddTV = fs.TVaddTV;
        fns.R = fs.R;
        fns.proj = fs.proj;
        fns.inpro = fs.inpro;
        fns.invTranv = fs.invTranv;
        fns.TranH = fs.TranH;
        fns.rank1operator = fs.rank1operator;
        fns.operadd = fs.operadd;
        if(params.StopCriterion == 4)
            fns.Tranv = fs.Tranv;
        end
    end
    % Check whether function handles are legal
    check_function_handle(fns, 'f');
    check_function_handle(fns, 'gf');
    check_function_handle(fns, 'Hv');
    check_function_handle(fns, 'STV');
    check_function_handle(fns, 'TVaddTV');
    check_function_handle(fns, 'R');
    check_function_handle(fns, 'proj');
    check_function_handle(fns, 'inpro');
    check_function_handle(fns, 'invTranv');
    check_function_handle(fns, 'TranH');
    check_function_handle(fns, 'rank1operator');
    check_function_handle(fns, 'operadd');
    if(params.StopCriterion == 4)
        check_function_handle(fns, 'Tranv');
    end
    if(params.StopCriterion == 0)
        check_function_handle(fns, 'IsStopped');
    end
    % Check whether required parameters are legal
    if(~isfield(params, 'x0') || ~isfield(params, 'B0'))
        error('Invalid arguments: missing initial x0 or B0');
    end

    fns.haveprec = 0;
    fns.g = fns.inpro;
    % Initialization
    c = params.c;
    tau1 = params.tau1;
    tau2 = params.tau2;
    err = inf;
    err_v = inf;
    err_x = inf;
    times = 0;
    Delta = 1;
    x = params.x0;
    err_g = 1;
    err_r = 1;
    [fx, x] = fns.f(x);
    [gradf, x] = fns.gf(x);
    B = params.B0;
    
    X{1} = x;
    F(1) = fx;
    G(1) = sqrt(fns.inpro(x, gradf, gradf));

    nf = 1;
    ng = 1;
    nH = 0;
    nV = 0;
    nR = 0;
    ind = 1;
    Grads = [];
    updateinfo = '';
    timeinfo = '';
    fprintf('times : %d, f(x) : %e\n', times, fx);
    tic
    T(1) = 0;
    while (err > params.error && times < params.max_t)
        if(params.debug > 1)
            tCGtimestart = toc;
        end
        EMPTYTV = fns.STV(x, 0, gradf);
        [eta, inner_it, stop_tCG, timeinfo] = tCG_TR(fns, x, gradf, EMPTYTV, Delta, times, params, timeinfo, 'SR1', B);
        if(params.debug > 1)
            timeinfo = sprintf('time, tCG:%.2e,', toc - tCGtimestart);
            getsytimestart = toc;
        end
        xtemp = fns.R(x, eta);
        [fxtemp, xtemp] = fns.f(xtemp);
        Hvtemp = fns.Hv(x, B, eta);
%         rho = (fx - fxtemp) / (- fns.inpro(x, gradf, eta) - 0.5 * fns.inpro(x, eta, Hvtemp));
        rho = (fx - fxtemp) / (- fns.inpro(x, fns.TVaddTV(x, gradf,  fns.STV(x, 0.5, Hvtemp)), eta));
        %update B
        s = eta;
        [gradftemp, xtemp] = fns.gf(xtemp);
        [iTgf, eta, x, xtemp] = fns.invTranv(x, eta, xtemp, gradftemp);
        if(params.debug > 2)
            cprintf([1,0,1], sprintf('\t\tCheck isometry of inverse vector transport: |gf| / |inv T gf|: %e/%e \n', ...
                sqrt(fns.inpro(xtemp, gradftemp, gradftemp)), sqrt(fns.inpro(x, iTgf, iTgf))));
        end
        y = fns.TVaddTV(x, iTgf, fns.STV(x, -1, gradf));
%         y = iTgf - gradf;
        if(params.debug > 1)
            timeinfo = [timeinfo sprintf('get s y f gf:%.2e,', toc - getsytimestart)];
            updateBtimestart = toc;
        end
        nV = nV + 1;
        Temp = fns.TVaddTV(x, y, fns.STV(x, -1, Hvtemp));
%         Temp = y - Hvtemp;
        denorm = fns.inpro(x, s, Temp);
        if(times == 0 && params.isconvex)
            B = (fns.inpro(x, y, y) / fns.inpro(x, s, y)) * B;
        end
        if(params.debug > 2)
            cprintf([1,0,1], sprintf('\t\tCheck isometry of Hessian approximation: g(eta, B eta) / g(B eta, eta): %e/%e \n', ...
                fns.inpro(x, s, fns.Hv(x, B, s)), fns.inpro(x, fns.Hv(x, B, s), s)));
        end
        if(abs(denorm) > sqrt(eps) * sqrt(fns.inpro(x, s, s) * fns.inpro(x, Temp, Temp)))
            B = fns.operadd(x, B, fns.rank1operator(x, fns.STV(x, 1 / denorm, Temp), Temp));
        else
            fprintf('warnning! denormator is close to zero.\n');
        end
        if(params.debug > 1)
            timeinfo = [timeinfo sprintf('update B:%.2e,', toc - updateBtimestart)];
        end
        if(rho > 0.75)
            if(stop_tCG == 2 || stop_tCG == 1)
                Delta = tau2 * Delta;
            end
            if(Delta > params.max_Delta)
                fprintf('reach max of Delta\n');
                break;
            end
        elseif(rho < 0.25)
            Delta = tau1 * Delta;
            if(Delta < params.min_Delta)
                fprintf('reach min Delta\n');
                break;
            end
        end
        if(rho > c || (abs((fx - fxtemp)/(abs(fx) + 1)) < sqrt(eps) && fxtemp < fx))
            B = fns.TranH(x, eta, xtemp, B);
            err_r = abs((fx - fxtemp) / fxtemp);
            if(params.StopCriterion == 4)
                if(params.debug > 1)
                    checkstoppartlysmooth = toc;
                end
                err_x = sqrt(fns.inpro(x, eta, eta));% norm(x - xtemp);%%---Whether err_x is OK is still a problem.
                                                      %% The ideal form should be dist(x1, x2) which may be expensive.
                for i = 1 : size(Grads, 2)
                    if(i ~= ind)
                        Gradsi.TV = reshape(Grads(:, i), size(gradf.TV));
                        [tempv, eta, x, xtemp] = fns.Tranv(x, eta, xtemp, Gradsi);
                        Grads(:, i) = reshape(tempv.TV, [], 1);
                        nV = nV + 1;
                    end
                end
                if(err_x < params.err_x)
                    Grads(:, ind) = reshape(gradftemp.TV, [], 1);
                    ind = ind + 1;
                    if(ind > params.num_Grads)
                        ind = ind - params.num_Grads;
                    end
                    [w, err_vectorv] = qpsubprob(Grads, 1);
                    err_vector.TV = reshape(err_vectorv, size(gradf.TV));
                    err_v = sqrt(fns.inpro(x, err_vector, err_vector));
                end
                if(params.debug > 1)
                    timeinfo = [timeinfo sprintf('check stop:%.2e,', toc - checkstoppartlysmooth)];
                end
            end
            x = xtemp;
            gradf = gradftemp;
            fx = fxtemp;
            err_g = sqrt(fns.inpro(x, gradf, gradf));
            if(params.debug > 0)
                updateinfo = '';
            end
        else
            if(params.debug > 0)
                updateinfo = ('without updating x_k');
            end
        end
        times = times + 1;
        nf = nf + 1;
        ng = ng + 1;
        nR = nR + 1;
        nH = nH + 1 + inner_it;
        X{times + 1} = x;
        if(times + 1 - params.keepnumX > 0)
            X{times + 1 - params.keepnumX} = [];
        end
        F(times + 1) = fx;
        G(times + 1) = err_g;
        T(times + 1) = toc;
        if(params.StopCriterion == 0)
            err = 1 - fns.IsStopped(X, F, G, T, params, fns);
        elseif(params.StopCriterion == 1)
            err = err_r;
        elseif(params.StopCriterion == 2)
            err = err_g;
        elseif(params.StopCriterion == 3)
            err = err_g / G(1);
        elseif(params.StopCriterion == 4)
            err = err_v;
        end
        if(params.debug > 1)
            cprintf('k', ['\t' timeinfo '\n']);
        end
        if(params.debug > 0)
            if(params.StopCriterion ~= 4)
                if(isempty(updateinfo))
                    fprintf('i:%d,f:%.3e,df/f:%.3e,|gf|:%.3e,radius:%.3e,time:%.2e,stoptCG:%d,in_it:%d,rho:%.2e\n', ...
                        times, fx, err_r, err_g, Delta, T(end), stop_tCG, inner_it, rho);
                else
                    fprintf(['i:%d, ' updateinfo ', radius reduces to %.3e\n'], times, Delta);
                end
            else
                if(isempty(updateinfo))
                    fprintf('i:%d,f:%.3e,df/f:%.3e,|gf|:%.3e,dx:%.3e,dg:%.3e,radius:%.3e,time:%.2e,stoptCG:%d,in_it:%d,rho:%.2e\n', ...
                        times, fx, err_r, err_g, err_x, err_v, Delta, T(end), stop_tCG, inner_it, rho);
                else
                    fprintf(['i:%d, ' updateinfo ', radius reduces to %.3e\n'], times, Delta);
                end
            end
        end
    end
    timecost = toc;
    fprintf('Num. of Iter.: %d, time cost: %f seconds, Num. of Fun, Gra, Hv, VT and R: %d, %d, %d, %d, %d \n', times, timecost, nf, ng, nH, nV, nR)
    fprintf('f: %e, |gradf|: %e, |gradf/gradf0|: %e \n', F(end), G(end), G(end)/G(1))
    if(params.isplot)
        figure(1);clf
        scatter(1 : length(F), F, '.');
        ylabel('f(x_i)');
        xlabel('iteration number');
        figure(2);clf
        semilogy(1 : length(G), G, '.');
        ylabel('|grad(x_i)|');
        xlabel('iteration number');
    end
    varargout{1} = X;
    varargout{2} = F;
    varargout{3} = G;
    varargout{4} = timecost;
    varargout{5} = nf;
    varargout{6} = ng;
    varargout{7} = nH;
    varargout{8} = T;
    varargout{9} = nV;
    varargout{10} = nR;
end
